/*
  pam_sfs.c
  Copyright 2004-2005 by Code Cobra


  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <errno.h>
#include <fcntl.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>


#define PAM_SM_AUTH
#define PAM_SM_SESSION
#include <security/pam_modules.h>

void safe_strcat(char*,const char*,int);

PAM_EXTERN int pam_sm_authenticate(pam_handle_t *pamh,int flags,int argc,const char **argv) {
  openlog("pam_sfs",0,LOG_AUTH);

  syslog(LOG_DEBUG,"%s","pam_sm_authenticate");

  //read the authentication information
  const char *user;
  const char *pass;

  if (pam_get_user(pamh,&user,NULL)!=PAM_SUCCESS)
    return PAM_SERVICE_ERR;

  if (pam_get_item(pamh,PAM_AUTHTOK,(const void**)&pass)!=PAM_SUCCESS)
    return PAM_SERVICE_ERR;

  //get the pw entry for the given user
  struct passwd *pw=getpwnam(user);
  if (pw==NULL)
    return PAM_USER_UNKNOWN;

  syslog(LOG_DEBUG,"getuid()=%d, user's id=%d",getuid(),pw->pw_uid);

  //check if user's home-directory matches /sfs/<host>/
  //if not, exit right away
  //if it does, extract <host>
  char *homedir=pw->pw_dir;

  char *sfs="/sfs/";

  if (strncmp(homedir,sfs,strlen(sfs))!=0)
    return PAM_SUCCESS; //home dir not in /sfs

  char *serverBegin=homedir+strlen(sfs);
  char *serverEnd=strchr(serverBegin,'/');

  //absolute max of 2048 characters in server name
  int lenServer=serverEnd-serverBegin+1;
  if (lenServer>2048)
    return PAM_SERVICE_ERR;

  char *server=malloc(lenServer);
  strncpy(server,serverBegin,lenServer-1);
  server[lenServer-1]='\0';

  int pid=fork();
  if (pid==0) {
    //child
    //change to the desired user
    if (setuid(pw->pw_uid)!=0) {
      syslog(LOG_ERR,"Unable to switch to uid.");
      exit(EXIT_FAILURE);
    } //setuid

    system("/usr/local/sfs-0.7.2/bin/sfsagent -c 2>/dev/null");

    //use a temporary variable to assemble the command in
    int maxCmd=4096;
    char *cmd=malloc(maxCmd);

    strcpy(cmd,"export HOME=/ ; /usr/local/sfs-0.7.2/bin/sfskey -p 0 add ");
    safe_strcat(cmd,user,maxCmd);
    safe_strcat(cmd,"@",maxCmd);
    safe_strcat(cmd,server,maxCmd);
    safe_strcat(cmd," 2>/dev/null",maxCmd);

    //syslog(LOG_DEBUG,"executing: %s",cmd);

    FILE *f=popen(cmd,"w");

    free(cmd);
    free(server); //child's copy

    if (f==NULL) {
      syslog(LOG_ERR,"Unable to start sfskey.");
      exit(EXIT_FAILURE);
    } //f

    fprintf(f,"%s",pass);

    int status=pclose(f);
    if (status!=0) {
      syslog(LOG_ERR,"Error from sfskey: %d.",status);
      exit(EXIT_FAILURE);
    } //pclose

    exit(EXIT_SUCCESS);
  } else {
    //parent, wait for child
    int status;
    waitpid(pid,&status,0);
    free(server); //parent's copy
  }

  closelog();

  return PAM_SUCCESS;
} //pam_sm_authenticate

PAM_EXTERN int pam_sm_setcred(pam_handle_t *pamh, int flags, int argc, const char **argv) {
  return PAM_SUCCESS;
} //pam_sm_setcred

PAM_EXTERN int pam_sm_open_session(pam_handle_t *pamh,int flags,int argc,const char **argv) {
  syslog(LOG_DEBUG,"pam_sm_open_session");
  return PAM_SUCCESS;
} //pam_sm_open_session

PAM_EXTERN int pam_sm_close_session(pam_handle_t *pamh,int flags,int argc,const char **argv) {
  syslog(LOG_DEBUG,"pam_sm_close_session");
  return PAM_SUCCESS;
} //pam_sm_close_session

void safe_strcat(char *dest,const char *src,int lenMax) {
  int lenCurrent=strlen(dest);
  int lenSource=strlen(src);

  if ((lenCurrent+lenSource)<lenMax)
    strcat(dest,src);
  else
    syslog(LOG_WARNING,"unsafe strcat");
} //safe_strcat

//end
